/* -------------------------------------------- */
import { TeDeumUtility } from "../common/tedeum-utility.js";
import { TeDeumRollDialog } from "../dialogs/tedeum-roll-dialog.js";

/* -------------------------------------------- */
/* -------------------------------------------- */
/**
 * Extend the base Actor entity by defining a custom roll data structure which is ideal for the Simple system.
 * @extends {Actor}
 */
export class TeDeumActor extends Actor {

  /* -------------------------------------------- */
  /**
   * Override the create() function to provide additional SoS functionality.
   *
   * This overrided create() function adds initial items
   * Namely: Basic skills, money,
   *
   * @param {Object} data        Barebones actor data which this function adds onto.
   * @param {Object} options     (Unused) Additional options which customize the creation workflow.
   *
   */

  static async create(data, options) {

    // Case of compendium global import
    if (data instanceof Array) {
      return super.create(data, options);
    }
    // If the created actor has items (only applicable to duplicated actors) bypass the new actor creation logic
    if (data.items) {
      let actor = super.create(data, options);
      return actor;
    }

    if (data.type == 'pj' ||  data.type == 'pnj') {
      const skills = await TeDeumUtility.loadCompendium("fvtt-te-deum.competences")
      data.items = data.items || []
      for (let skill of skills) {
        if (skill.system.isBase || skill.system.score == 1) {
          data.items.push(skill.toObject())
        }
      }
    }

    return super.create(data, options);
  }

  /* -------------------------------------------- */
  async prepareData() {
    super.prepareData()


  }

  /* -------------------------------------------- */
  prepareDerivedData() {
    super.prepareDerivedData();
  }

  /* -------------------------------------------- */
  _preUpdate(changed, options, user) {

    super._preUpdate(changed, options, user);
  }

  getCompetenceScore(compName) {
    let competence = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == compName.toLowerCase())
    if (competence) {
      return competence.system.score
    }
    return 0
  }

  /* -------------------------------------------- */
  _onUpdate(changed, options, userId) {
    let updates = []
    let memoriser = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "mémoriser")
    let newScore = this.getCommonBaseValue(this.system.caracteristiques.adresse.value)
    if (memoriser && memoriser?.system.score != newScore) {
      updates.push({ _id: memoriser.id, "system.score": Number(newScore) })
    }

    let perception = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "perception")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.sensibilite.value)
    if (perception && perception.system.score != newScore) {
      updates.push({ _id: perception.id, "system.score": Number(newScore) })
    }

    let charme = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "charme")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.entregent.value)
    if (charme && charme?.system.score != newScore) {
      updates.push({ _id: charme.id, "system.score": Number(newScore) })
    }

    let endurance = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "endurance")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.complexion.value)
    if (endurance && endurance?.system.score != newScore) {
      updates.push({ _id: endurance.id, "system.score": Number(newScore) })
    }

    let course = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "course")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.adresse.value)
    if (course && course?.system.score != newScore) {
      updates.push({ _id: course.id, "system.score": Number(newScore) })
    }

    let initiative = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "initiative")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.adresse.value)
    if (initiative && initiative?.system.score != newScore) {
      updates.push({ _id: initiative.id, "system.score": Number(newScore) })
    }

    let actionsTour = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "actions/tour")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.adresse.value)
    if (actionsTour && actionsTour?.system.score != newScore) {
      updates.push({ _id: actionsTour.id, "system.score": Number(newScore) })
    }

    let effort = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "effort")
    newScore = this.getCommonBaseValue(this.system.caracteristiques.puissance.value)
    if (effort && effort?.system.score != newScore) {
      updates.push({ _id: effort.id, "system.score": Number(newScore) })
    }

    if (updates.length > 0) {
      this.updateEmbeddedDocuments('Item', updates)
    }

    super._onUpdate(changed, options, userId);
  }

  /* -------------------------------------------- */
  async _preCreate(data, options, user) {
    await super._preCreate(data, options, user);

    // Configure prototype token settings
    const prototypeToken = {};
    if (this.type === "pj") Object.assign(prototypeToken, {
      sight: { enabled: true }, actorLink: true, disposition: CONST.TOKEN_DISPOSITIONS.FRIENDLY
    });
    this.updateSource({ prototypeToken });
  }

  /* -------------------------------------------- */
  getCommonBaseValue(value) {
    return game.system.tedeum.config.COMMON_VALUE[value]?.value || 0
  }
  getInitiative() {
    return game.system.tedeum.config.COMMON_VALUE[this.system.caracteristiques.adresse.value]?.value || 0
  }
  /* -------------------------------------------- */
  getBonusDegats() {
    return game.system.tedeum.config.BONUS_DEGATS[this.system.caracteristiques.puissance.value]
  }
  /* -------------------------------------------- */
  getNbArmures() {
    return game.system.tedeum.config.MAX_ARMURES_LOURDES[this.system.caracteristiques.puissance.value]
  }
  getNbActions() {
    return game.system.tedeum.config.ACTIONS_PAR_TOUR[this.system.caracteristiques.adresse.value]
  }
  getInitiative() {
    return game.system.tedeum.config.ACTIONS_PAR_TOUR[this.system.caracteristiques.adresse.value]
  }
  getNbArmuresLourdesActuel() {
    let armures = this.getArmures()
    let nb = 0
    for (let armure of armures) {
      if (armure.system.equipe) {
        nb += armure.system.coutArmureLourde
      }
    }
    return nb
  }
  /* -------------------------------------------- */
  getEducations() {
    let educations = this.items.filter(item => item.type == 'education')
    return educations
  }
  /* -------------------------------------------- */
  getCompetences() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'competence') || [])
    return comp;
  }
  /* -------------------------------------------- */
  getGraces() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'grace') || [])
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getArmes() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'arme') || [])
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getEquipements() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'equipement') || [])
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getSimples() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'simple') || [])
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getArmures() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'armure') || [])
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getBlessures() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'blessure') || [])
    for (let c of comp) {
      let blessDef = game.system.tedeum.config.blessures[c.system.typeBlessure]
      c.malus = blessDef.modifier
    }
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getMaladies() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'maladie') || [])
    for (let c of comp) {
      c.malus = "N/A"
      if (c.system.appliquee) {
        let malDef = game.system.tedeum.config.virulence[c.system.virulence]
        c.malus = malDef.modifier
      }
    }
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  getPoisons() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'poison') || [])
    for (let c of comp) {
      c.malus = "N/A"
      if (c.system.appliquee) {
        let poisDef = game.system.tedeum.config.virulencePoison[c.system.virulence]
        c.malus = poisDef.modifier
      }
    }
    TeDeumUtility.sortArrayObjectsByName(comp)
    return comp;
  }
  /* -------------------------------------------- */
  getSanteModifier() {
    let comp = foundry.utils.duplicate(this.items.filter(item => item.type == 'maladie') || [])
    let modTotal = 0
    for (let c of comp) {
      if (c.system.appliquee) {
        let maladieDef = game.system.tedeum.config.virulence[c.system.virulence]
        modTotal += maladieDef.modifier
      }
    }
    let simples = foundry.utils.duplicate(this.items.filter(item => item.type == 'simple') || [])
    for (let c of simples) {
      if (c.system.appliquee) {
        let simpleDef = game.system.tedeum.config.virulencePoison[c.system.virulence]
        modTotal += simpleDef.modifier
      }
    }
    let blessures = foundry.utils.duplicate(this.items.filter(item => item.type == 'blessure') || [])
    for (let c of blessures) {
      let blessDef = game.system.tedeum.config.blessures[c.system.typeBlessure]
      modTotal += blessDef.modifier
    }
    // Si le nombre de blessures est supérieur au score d'endurance, alors malus supplémentaire
    let endurance = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "endurance")
    if (blessures.length > endurance.system.score) {
      modTotal += -1
    }
    return modTotal
  }
  /* -------------------------------------------- */
  async appliquerDegats(rollData) {
    let combat = this.prepareCombat()
    rollData.defenderName = this.name
    let touche = combat[rollData.loc.id].touche
    if (rollData.degats > 0 && rollData.degats > touche) {
      let diff = rollData.degats - touche
      for (let bId in game.system.tedeum.config.blessures) {
        let blessure = game.system.tedeum.config.blessures[bId]
        if (diff >= blessure.degatsMin && diff <= blessure.degatsMax) {
          // Create a new blessure object
          let blessureObj = {
            name: blessure.label,
            type: "blessure",
            system: {
              typeBlessure: bId,
              localisation: rollData.loc.id,
              appliquee: true,
              description: "Blessure infligée par un coup de " + rollData.arme.name + " de " + rollData.alias,
            }
          }
          rollData.blessure = blessureObj
          this.createEmbeddedDocuments('Item', [blessureObj]);
        }
      }
    }
    // Display the relevant chat message
    let msg = await TeDeumUtility.createChatWithRollMode(rollData.alias, {
      content: await renderTemplate(`systems/fvtt-te-deum/templates/chat/chat-blessure-result.hbs`, rollData)
    })
    await msg.setFlag("world", "te-deum-rolldata", rollData)
  }

  /* -------------------------------------------- */
  updateCarac(c, key) {
    c.key = key
    c.name = game.system.tedeum.config.caracteristiques[key].label
    c.generalqualite = game.system.tedeum.config.descriptionValeur[c.value].qualite
    if (this.system.genre.toLowerCase() == "homme") {
      c.qualite = game.system.tedeum.config.descriptionValeur[c.value][key]
    } else {
      c.qualite = game.system.tedeum.config.descriptionValeurFemme[c.value][key]
    }
    c.dice = game.system.tedeum.config.descriptionValeur[c.value].dice
    c.negativeDice = game.system.tedeum.config.descriptionValeur[c.value].negativeDice
  }

  /* -------------------------------------------- */
  prepareCaracteristiques() {
    let carac = foundry.utils.deepClone(this.system.caracteristiques)
    for (let key in carac) {
      let c = carac[key]
      this.updateCarac(c, key)
      c.description = game.system.tedeum.config.caracteristiques[key].description
    }
    return carac
  }
  /* -------------------------------------------- */
  prepareProvidence() {
    let providence = foundry.utils.deepClone(this.system.providence)
    providence.name = "Providence"
    if (this.system.genre.toLowerCase() == "homme") {
      providence.qualite = game.system.tedeum.config.providence[providence.value].labelH
    } else {
      providence.qualite = game.system.tedeum.config.providence[providence.value].labelF
    }
    providence.dice = game.system.tedeum.config.providence[providence.value].diceValue
    providence.description = "La Providence représente la Volonté Divine à l'œuvre pour guider ou sauver un être humain. Les PJ montent dans l’échelle de la Providence en menant à bien leurs missions et en se montrant vertueux. Les points de Providence peuvent servir à augmenter temporairement une caractéris- tique, à modifier la gravité d'une blessure, et à résister au vieillissement. Chaque person- nage commence avec un score initial de 1 en Providence (au niveau Pauvre pécheur)."
    return providence
  }
  /* -------------------------------------------- */
  prepareCombat() {
    let combatLoc = foundry.utils.deepClone(this.system.localisation)
    for (let key in combatLoc) {
      combatLoc[key] = foundry.utils.mergeObject(combatLoc[key], game.system.tedeum.config.LOCALISATION[key])
      combatLoc[key].armures = []
      combatLoc[key].blessures = []
      combatLoc[key].protectionTotal = 0
      let armures = this.getArmures()
      for (let armure of armures) {
        if (armure.system.equipe && armure.system.localisation[key].protege) {
          combatLoc[key].armures.push(armure)
          combatLoc[key].protectionTotal += armure.system.protection
        }
      }
      let blessures = this.getBlessures()
      for (let blessure of blessures) {
        if (blessure.system.localisation == key) {
          combatLoc[key].blessures.push(blessure)
        }
      }
      let endurance = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == "endurance")
      combatLoc[key].endurance = endurance.system.score + game.system.tedeum.config.LOCALISATION[key].locMod
      combatLoc[key].touche = combatLoc[key].endurance + combatLoc[key].protectionTotal
    }
    return combatLoc
  }
  /* -------------------------------------------- */
  modifyProvidence(value) {
    let providence = foundry.utils.duplicate(this.system.providence)
    providence.value = Math.min(Math.max(providence.value + value, 0), 6)
    this.update({ "system.providence": providence })
  }

  /* -------------------------------------------- */
  async modifyXP(key, value) {
    let xp = this.system.caracteristiques[key].experience
    xp = Math.max(xp + value, 0)
    await this.update({ [`system.caracteristiques.${key}.experience`]: xp })
    this.sheet?.render(true)
  }

  /* -------------------------------------------- */
  filterCompetencesByCarac(key) {
    let comp = this.items.filter(item => item.type == 'competence' && item.system.caracteristique == key)
    comp.forEach(c => {
      if (c.system.isBase) {
        c.system.score = this.system.caracteristiques[c.system.caracteristique].value
      }
      let caracDice = game.system.tedeum.config.descriptionValeur[this.system.caracteristiques[c.system.caracteristique].value].dice
      c.system.formula = caracDice + "+" + c.system.score
    })
    comp = comp.sort((a, b) => a.name.localeCompare(b.name))
    return foundry.utils.deepClone(comp || {})
  }

  /* -------------------------------------------- */
  prepareArbreCompetences() {
    let arbre = foundry.utils.deepClone(this.system.caracteristiques)
    for (let key in arbre) {
      let c = arbre[key]
      this.updateCarac(c, key)
      c.competences = this.filterCompetencesByCarac(key)
    }
    return arbre
  }

  /* -------------------------------------------- */
  getItemById(id) {
    let item = this.items.find(item => item.id == id);
    if (item) {
      item = foundry.utils.duplicate(item)
    }
    return item;
  }

  /* -------------------------------------------- */
  async equipItem(itemId) {
    let item = this.items.find(item => item.id == itemId)
    if (!this.checkArmure(item)) {
      return
    }
    let update = { _id: item.id, "system.equipe": !item.system.equipe };
    await this.updateEmbeddedDocuments('Item', [update]); // Updates one EmbeddedEntity
  }

  /* ------------------------------------------- */
  checkArmure(item) {
    if (item.type != "armure") {
      return true
    }
    if (item.system.equipe) {
      return true
    }
    let nbArmuresLourdes = this.getNbArmuresLourdesActuel()
    if (nbArmuresLourdes + item.system.coutArmureLourde > this.getNbArmures().value) {
      ui.notifications.warn("Impossible d'équiper cette armure, nombre d'armures lourdes maximum atteint")
      return false
    }

    // Loop thru localisation
    let armures = this.getArmures()
    for (let loc in item.system.localisation) {
      if (item.system.localisation[loc].protege) {
        for (let armure of armures) {
          if (armure.system.equipe && armure.system.localisation[loc].protege) {
            let flag = true
            //console.log("Check armure", armure, item)=
            if (item.system.typeArmure == "cuir") {
              flag = armure.system.superposableCuir
            }
            if (item.system.typeArmure == "maille") {
              flag =  armure.system.superposableMaille
            }
            if (item.system.typeArmure == "plate") {
              flag = armure.system.superposablePlate
            }
            if (!flag) {
              ui.notifications.warn("Impossible d'équiper cette armure, non superposable")
              return flag
            }
          }
        }
      }
      return true
    }
  }

  /* ------------------------------------------- */
  async buildContainerTree() {
    let equipments = foundry.utils.duplicate(this.items.filter(item => item.type == "equipment") || [])
    for (let equip1 of equipments) {
      if (equip1.system.iscontainer) {
        equip1.system.contents = []
        equip1.system.contentsEnc = 0
        for (let equip2 of equipments) {
          if (equip1._id != equip2.id && equip2.system.containerid == equip1.id) {
            equip1.system.contents.push(equip2)
            let q = equip2.system.quantity ?? 1
            equip1.system.contentsEnc += q * equip2.system.weight
          }
        }
      }
    }

    // Compute whole enc
    let enc = 0
    for (let item of equipments) {
      //item.data.idrDice = TeDeumUtility.getDiceFromLevel(Number(item.data.idr))
      if (item.system.equipped) {
        if (item.system.iscontainer) {
          enc += item.system.contentsEnc
        } else if (item.system.containerid == "") {
          let q = item.system.quantity ?? 1
          enc += q * item.system.weight
        }
      }
    }
    for (let item of this.items) { // Process items/shields/armors
      if ((item.type == "weapon" || item.type == "shield" || item.type == "armor") && item.system.equipped) {
        let q = item.system.quantity ?? 1
        enc += q * item.system.weight
      }
    }

    // Store local values
    this.encCurrent = enc
    this.containersTree = equipments.filter(item => item.system.containerid == "") // Returns the root of equipements without container

  }

  /* -------------------------------------------- */
  async equipGear(equipmentId) {
    let item = this.items.find(item => item.id == equipmentId);
    if (item?.system) {
      let update = { _id: item.id, "system.equipped": !item.system.equipped };
      await this.updateEmbeddedDocuments('Item', [update]); // Updates one EmbeddedEntity
    }
  }

  /* -------------------------------------------- */
  clearInitiative() {
    this.getFlag("world", "initiative", -1)
  }

  /* -------------------------------------------- */
  getInitiativeScore() {
    let initiative = this.items.find(it => it.type == "competence" && it.name.toLowerCase() == "initiative")
    if (initiative) {
      return initiative.system.score
    }
    ui.notifications.warn("Impossible de trouver la compétence Initiative pour l'acteur " + this.name)
    return -1;
  }

  /* -------------------------------------------- */
  async deleteAllItemsByType(itemType) {
    let items = this.items.filter(item => item.type == itemType);
    await this.deleteEmbeddedDocuments('Item', items);
  }

  /* -------------------------------------------- */
  async addItemWithoutDuplicate(newItem) {
    let item = this.items.find(item => item.type == newItem.type && item.name.toLowerCase() == newItem.name.toLowerCase())
    if (!item) {
      await this.createEmbeddedDocuments('Item', [newItem]);
    }
  }

  /* -------------------------------------------- */
  async incDecQuantity(objetId, incDec = 0) {
    let objetQ = this.items.get(objetId)
    if (objetQ) {
      let newQ = objetQ.system.quantity + incDec
      if (newQ >= 0) {
        await this.updateEmbeddedDocuments('Item', [{ _id: objetQ.id, 'system.quantity': newQ }]) // pdates one EmbeddedEntity
      }
    }
  }

  /* -------------------------------------------- */
  getCommonRollData() {
    let rollData = TeDeumUtility.getBasicRollData()
    rollData.alias = this.name
    rollData.actorImg = this.img
    rollData.actorId = this.id
    rollData.img = this.img
    rollData.providence = this.prepareProvidence()
    rollData.santeModifier = this.getSanteModifier()

    return rollData
  }

  /* -------------------------------------------- */
  getCommonCompetence(compId) {
    let rollData = this.getCommonRollData()

    let competence = foundry.utils.duplicate(this.items.find(it => it.type == "competence" && it.id == compId))

    rollData.competence = competence
    let c = foundry.utils.duplicate(this.system.caracteristiques[competence.system.caracteristique])
    this.updateCarac(c, competence.system.caracteristique)
    rollData.carac = c
    rollData.img = competence.img

    return rollData
  }

  /* -------------------------------------------- */
  rollCompetence(compId) {
    let rollData = this.getCommonCompetence(compId)
    rollData.mode = "competence"
    rollData.title = rollData.competence.name
    this.startRoll(rollData).catch("Error on startRoll")
  }

  /* -------------------------------------------- */
  async rollDegatsArme(armeId) {
    let weapon = this.items.get(armeId)
    if (weapon) {
      let bDegats = 0
      if ( weapon.system.typeArme == "melee" ) {
        bDegats = this.getBonusDegats()
      }
      let formula = weapon.system.degats + "+" + bDegats.value
      let degatsRoll = await new Roll(formula).roll()
      await TeDeumUtility.showDiceSoNice(degatsRoll, game.settings.get("core", "rollMode") )
      let rollData = this.getCommonRollData()
      rollData.mode = "degats"
      rollData.formula = formula
      rollData.arme = foundry.utils.duplicate(weapon)
      rollData.degatsRoll = foundry.utils.duplicate(degatsRoll)
      rollData.degats = degatsRoll.total

      let msg = await TeDeumUtility.createChatWithRollMode(rollData.alias, {
        content: await renderTemplate(`systems/fvtt-te-deum/templates/chat/chat-degats-result.hbs`, rollData)
      })
      await msg.setFlag("world", "te-deum-rolldata", rollData)
      console.log("Rolldata result", rollData)
    }
  }

  /* -------------------------------------------- */
  rollArme(armeId, compName = undefined) {
    let weapon = this.items.get(armeId)
    if (weapon) {
      weapon = foundry.utils.duplicate(weapon)
      let rollData = this.getCommonRollData()
      rollData.mode = "arme"
      rollData.isTir = weapon.system.typeArme == "tir"
      rollData.arme = weapon
      rollData.img = weapon.img
      rollData.title = weapon.name
      rollData.porteeTir = "courte"
      rollData.porteeLabel = game.system.tedeum.config.ARME_PORTEES.courte.label
      rollData.isViser = false
      rollData.isMouvement = false

      // Display warning if not target defined
      if (!rollData.defenderTokenId) {
        ui.notifications.warn("Vous attaquez avec une arme : afin de bénéficier des automatisations, il est conseillé de selectionner une cible")
      }

      // Setup competence + carac
      if (!compName) {
        let compIdx = weapon.system.competence
        compName = game.system.tedeum.config.armeCompetences[compIdx]?.label
      }
      let competence = this.items.find(item => item.type == "competence" && item.name.toLowerCase() == compName.toLowerCase())
      if (competence) {
        rollData.competence = competence
        let c = foundry.utils.duplicate(this.system.caracteristiques[competence.system.caracteristique])
        this.updateCarac(c, competence.system.caracteristique)
        rollData.carac = c
      } else {
        ui.notifications.warn("Impossible de trouver la compétence " + compName)
        return
      }
      this.startRoll(rollData).catch("Error on startRoll")
    } else {
      ui.notifications.warn("Impossible de trouver l'arme concernée ")
    }
  }

  /* -------------------------------------------- */
  async startRoll(rollData) {
    console.log("startRoll", rollData)
    let rollDialog = await TeDeumRollDialog.create(this, rollData)
    rollDialog.render(true)
  }

}
